==> v0.13.11.tar.gz: git-next/crates/cli/src/tui/actor/model.rs <==
//
use ratatui::{
    layout::Alignment,
    prelude::{Buffer, Rect},
    style::{Color, Style, Stylize as _},
    symbols::border,
    text::{Line, Span},
    widgets::{
        block::{Position, Title},
        Block, Paragraph, StatefulWidget, Widget,
    },
};

use git_next_core::{
    git::{self, graph::Log, Commit},
    ForgeAlias, RepoAlias, RepoBranches,
};
use tracing::info;
use tui_scrollview::ScrollViewState;

use std::{collections::BTreeMap, fmt::Display, time::Instant};

use crate::{server::actor::messages::ValidAppConfig, tui::components::ConfiguredAppWidget};

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct State {
    last_update: Instant,
    started: Instant,
    pub mode: ServerState,
}
impl State {
    pub fn initial() -> Self {
        Self {
            last_update: Instant::now(),
            started: Instant::now(),
            mode: ServerState::Initial { tick: 0 },
        }
    }

    pub fn tap(&mut self) {
        self.last_update = Instant::now();
        if let ServerState::Initial { tick } = &mut self.mode {
            *tick += 1;
        }
    }

    fn beating_heart(&self) -> String {
        if self.last_update.duration_since(self.started).as_secs() % 2 == 0 {
            "💚 "
        } else {
            " 💚"
        }
        .to_string()
    }
}

fn time() -> String {
    chrono::Local::now().format("%H:%M").to_string()
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ServerState {
    /// UI has started but has no information on the state of the server
    Initial { tick: usize }, // NOTE: for use with throbber-widgets-tui ?

    /// The application configuration has been loaded, individual forges and repos have their own
    /// states
    Configured {
        forges: BTreeMap<ForgeAlias, ForgeState>,
    },
}
impl ServerState {
    pub fn update_branches(
        &mut self,
        forge_alias: &ForgeAlias,
        repo_alias: &RepoAlias,
        branches: RepoBranches,
    ) {
        if let Self::Configured { forges } = self {
            let Some(forge_state) = forges.get_mut(forge_alias) else {
                return;
            };
            let Some(repo_state) = forge_state.repos.get_mut(repo_alias) else {
                return;
            };
            match repo_state {
                RepoState::Configured {
                    branches: state_branches,
                    ..
                }
                | RepoState::Ready {
                    branches: state_branches,
                    ..
                } => *state_branches = branches,

                RepoState::Identified { .. } => (),
            }
        }
    }

    pub fn update_log(&mut self, forge_alias: &ForgeAlias, repo_alias: &RepoAlias, log: Log) {
        if let Self::Configured { forges } = self {
            let Some(forge_state) = forges.get_mut(forge_alias) else {
                return;
            };
            let Some(repo_state) = forge_state.repos.get_mut(repo_alias) else {
                return;
            };
            match repo_state {
                RepoState::Ready { log: state_log, .. } => *state_log = log,

                RepoState::Identified { .. } | RepoState::Configured { .. } => (),
            }
        }
    }
}
impl From<ValidAppConfig> for ServerState {
    fn from(app_config: ValidAppConfig) -> Self {
        Self::Configured {
            forges: app_config
                .app_config
                .forges()
                .map(|(forge_alias, config)| {
                    (
                        forge_alias,
                        config
                            .repos()
                            .map(|(repo_alias, server_repo_config)| {
                                (repo_alias, server_repo_config.repo_config())
                            })
                            .map(
                                |(repo_alias, option_repo_config)| match option_repo_config {
                                    Some(rc) => (
                                        repo_alias.clone(),
                                        RepoState::Configured {
                                            repo_alias,
                                            message: RepoMessage::builder()
                                                .text("configured".into())
                                                .style(Style::default().fg(Color::LightGreen))
                                                .build(),
                                            alert: None,
                                            branches: rc.branches().clone(),
                                            log: git::graph::Log::default(),
                                        },
                                    ),
                                    None => (
                                        repo_alias.clone(),
                                        RepoState::Identified {
                                            repo_alias,
                                            message: RepoMessage::builder()
                                                .text("identified".into())
                                                .style(Style::default().fg(Color::Gray))
                                                .build(),
                                            alert: None,
                                        },
                                    ),
                                },
                            )
                            .collect::<Vec<_>>(),
                    )
                })
                .map(|(forge_alias, vec_repo_alias_state)| {
                    let forge_state: ForgeState = ForgeState {
                        alias: forge_alias.clone(),
                        view_state: ViewState::default(),
                        repos: vec_repo_alias_state.into_iter().collect::<BTreeMap<_, _>>(),
                    };
                    (forge_alias, forge_state)
                })
                .collect::<BTreeMap<_, _>>(),
        }
    }
}

#[derive(Copy, Clone, Debug, Default, PartialEq, Eq)]
pub enum ViewState {
    Collapsed,
    #[default]
    Expanded,
}
impl Display for ViewState {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let view_state = match self {
            Self::Collapsed => "+",
            Self::Expanded => "-",
        };
        write!(f, "{view_state}")
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct ForgeState {
    pub alias: ForgeAlias,
    pub view_state: ViewState,
    pub repos: BTreeMap<RepoAlias, RepoState>,
}

#[bon::builder]
#[derive(Clone, Debug, PartialEq, Eq)]
pub struct RepoMessage {
    text: String,
    style: Style,
}
impl From<&RepoMessage> for Span<'_> {
    fn from(value: &RepoMessage) -> Self {
        Self::default()
            .content(value.text.clone())
            .style(value.style)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum RepoState {
    Identified {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
    },
    Configured {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
        branches: RepoBranches,
        log: Log,
    },
    Ready {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
        branches: RepoBranches,
        view_state: ViewState,
        main: Commit,
        next: Commit,
        dev: Commit,
        log: Log,
    },
}
impl RepoState {
    #[tracing::instrument]
    pub fn update_branches(&mut self, branches: RepoBranches) {
        match self {
            Self::Configured {
                branches: state_branches,
                ..
            }
            | Self::Ready {
                branches: state_branches,
                ..
            } => {
                *state_branches = branches;
            }

            Self::Identified { .. } => (),
        }
    }

    #[tracing::instrument]
    pub fn update_log(&mut self, log: Log) {
        match self {
            Self::Configured { log: state_log, .. } | Self::Ready { log: state_log, .. } => {
                *state_log = log;
            }

            Self::Identified { .. } => {
                info!("git graph log ignored by ui");
            }
        }
    }

    #[tracing::instrument]
    pub fn update_message(&mut self, msg: impl Into<String> + std::fmt::Debug, colour: Color) {
        match self {
            Self::Identified { message, .. }
            | Self::Configured { message, .. }
            | Self::Ready { message, .. } => {
                info!(?msg, "updating ui");
                *message = RepoMessage::builder()
                    .text(msg.into())
                    .style(Style::default().fg(colour))
                    .build();
            }
        }
    }

    #[tracing::instrument]
    pub fn clear_alert(&mut self) {
        match self {
            Self::Identified { alert, .. }
            | Self::Configured { alert, .. }
            | Self::Ready { alert, .. } => {
                *alert = None;
            }
        }
    }

    #[tracing::instrument]
    pub fn alert(&mut self, msg: impl Into<String> + std::fmt::Debug) {
        let msg: String = msg.into();
        tracing::info!(%msg, "new tui alert");
        self.update_message("ALERT", Color::Red);
        match self {
            Self::Identified { alert, .. }
            | Self::Configured { alert, .. }
            | Self::Ready { alert, .. } => *alert = Some(msg),
        }
    }

    pub fn ready(self, main: Commit, next: Commit, dev: Commit) -> Self {
        match self {
            Self::Identified {
                repo_alias,
                message,
                alert,
            } => Self::Identified {
                repo_alias,
                message,
                alert,
            },
            Self::Configured {
                repo_alias,
                message,
                alert,
                branches,
                log,
            } => Self::Ready {
                repo_alias,
                message,
                alert,
                branches,
                view_state: ViewState::Expanded,
                main,
                next,
                dev,
                log,
            },
            Self::Ready {
                repo_alias,
                message,
                alert,
                branches,
                view_state,
                log,
                .. // drop existing main, next and dev to use parameters
            } => Self::Ready {
                repo_alias,
                message,
                alert,
                branches,
                view_state,
                main,
                next,
                dev,
                log,
            },
        }
    }
}

impl StatefulWidget for &State {
    type State = ScrollViewState;
    fn render(self, area: Rect, buf: &mut Buffer, state: &mut Self::State)
    where
        Self: Sized,
    {
        let block = Block::bordered()
            .title(
                Title::from(format!(" Git-Next v{} ", clap::crate_version!()).bold())
                    .alignment(Alignment::Center),
            )
            .title(
                Title::from(Line::from(vec![
                    " [q]uit ".into(),
                    self.beating_heart().into(),
                    " ".into(),
                ]))
                .alignment(Alignment::Center)
                .position(Position::Bottom),
            )
            .title(
                Title::from(format!(" {} ", time()))
                    .alignment(Alignment::Right)
                    .position(Position::Bottom),
            )
            .border_set(border::THICK);
        let interior = block.inner(area);
        block.render(area, buf);
        match &self.mode {
            ServerState::Initial { tick } => Paragraph::new(format!("Loading...{tick}"))
                .centered()
                .render(interior, buf),
            ServerState::Configured { forges } => {
                ConfiguredAppWidget { forges }.render(interior, buf, state);
            }
        }
    }
}

==> v0.14.1.tar.gz: git-next/crates/cli/src/tui/actor/model.rs <==
//
use std::{
    collections::BTreeMap,
    fmt::{Debug, Display},
    time::Instant,
};

use derive_more::derive::{DerefMut, Display};
use ratatui::{
    layout::Alignment,
    prelude::{Buffer, Rect},
    style::{Color, Style, Stylize as _},
    symbols::border,
    text::{Line, Span},
    widgets::{Block, Paragraph, StatefulWidget, Widget},
};
use tracing::info;
use tui_scrollview::ScrollViewState;

use git_next_core::{
    git::{self, graph::Log, Commit},
    newtype, s, ForgeAlias, RepoAlias, RepoBranches,
};

use crate::{
    server::actor::messages::ValidAppConfig,
    tui::components::{key_focus::KeyFocus, ConfiguredAppWidget},
};

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct State {
    last_update: Instant,
    started: Instant,
    pub mode: ServerState,
    pub filter: UIRepoFilter,
    pub key_focus: KeyFocus,
}
impl State {
    pub fn initial() -> Self {
        Self {
            last_update: Instant::now(),
            started: Instant::now(),
            mode: ServerState::Initial { tick: 0 },
            filter: UIRepoFilter::default(),
            key_focus: KeyFocus::default(),
        }
    }

    pub fn tap(&mut self) {
        self.last_update = Instant::now();
        if let ServerState::Initial { tick } = &mut self.mode {
            *tick += 1;
        }
    }

    fn beating_heart(&self) -> String {
        if self.last_update.duration_since(self.started).as_secs() % 2 == 0 {
            "💚 "
        } else {
            " 💚"
        }
        .to_string()
    }
}

fn time() -> String {
    chrono::Local::now().format("%H:%M").to_string()
}

newtype!(
    UIRepoFilter,
    String,
    Default,
    DerefMut,
    Display,
    "Filter to limit repos shown"
);

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ServerState {
    /// UI has started but has no information on the state of the server
    Initial { tick: usize }, // NOTE: for use with throbber-widgets-tui ?

    /// The application configuration has been loaded, individual forges and repos have their own
    /// states
    Configured {
        forges: BTreeMap<ForgeAlias, ForgeState>,
    },
}
impl ServerState {
    pub fn update_branches(
        &mut self,
        forge_alias: &ForgeAlias,
        repo_alias: &RepoAlias,
        branches: RepoBranches,
    ) {
        if let Self::Configured { forges } = self {
            let Some(forge_state) = forges.get_mut(forge_alias) else {
                return;
            };
            let Some(repo_state) = forge_state.repos.get_mut(repo_alias) else {
                return;
            };
            match repo_state {
                RepoState::Configured {
                    branches: state_branches,
                    ..
                }
                | RepoState::Ready {
                    branches: state_branches,
                    ..
                } => *state_branches = branches,

                RepoState::Identified { .. } => (),
            }
        }
    }

    pub fn update_log(&mut self, forge_alias: &ForgeAlias, repo_alias: &RepoAlias, log: Log) {
        if let Self::Configured { forges } = self {
            let Some(forge_state) = forges.get_mut(forge_alias) else {
                return;
            };
            let Some(repo_state) = forge_state.repos.get_mut(repo_alias) else {
                return;
            };
            match repo_state {
                RepoState::Ready { log: state_log, .. } => *state_log = log,

                RepoState::Identified { .. } | RepoState::Configured { .. } => (),
            }
        }
    }
}
impl From<ValidAppConfig> for ServerState {
    fn from(app_config: ValidAppConfig) -> Self {
        Self::Configured {
            forges: app_config
                .app_config
                .forges()
                .map(|(forge_alias, config)| {
                    (
                        forge_alias,
                        config
                            .repos()
                            .map(|(repo_alias, server_repo_config)| {
                                (repo_alias, server_repo_config.repo_config())
                            })
                            .map(
                                |(repo_alias, option_repo_config)| match option_repo_config {
                                    Some(rc) => (
                                        repo_alias.clone(),
                                        RepoState::Configured {
                                            repo_alias,
                                            message: RepoMessage::builder()
                                                .text("configured".into())
                                                .style(Style::default().fg(Color::LightGreen))
                                                .build(),
                                            alert: None,
                                            branches: rc.branches().clone(),
                                            log: git::graph::Log::default(),
                                        },
                                    ),
                                    None => (
                                        repo_alias.clone(),
                                        RepoState::Identified {
                                            repo_alias,
                                            message: RepoMessage::builder()
                                                .text("identified".into())
                                                .style(Style::default().fg(Color::Gray))
                                                .build(),
                                            alert: None,
                                        },
                                    ),
                                },
                            )
                            .collect::<Vec<_>>(),
                    )
                })
                .map(|(forge_alias, vec_repo_alias_state)| {
                    let forge_state: ForgeState = ForgeState {
                        alias: forge_alias.clone(),
                        view_state: ViewState::default(),
                        repos: vec_repo_alias_state.into_iter().collect::<BTreeMap<_, _>>(),
                    };
                    (forge_alias, forge_state)
                })
                .collect::<BTreeMap<_, _>>(),
        }
    }
}

#[derive(Copy, Clone, Debug, Default, PartialEq, Eq)]
pub enum ViewState {
    Collapsed,
    #[default]
    Expanded,
}
impl Display for ViewState {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let view_state = match self {
            Self::Collapsed => "+",
            Self::Expanded => "-",
        };
        write!(f, "{view_state}")
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct ForgeState {
    pub alias: ForgeAlias,
    pub view_state: ViewState,
    pub repos: BTreeMap<RepoAlias, RepoState>,
}

#[derive(bon::Builder, Clone, Debug, PartialEq, Eq)]
pub struct RepoMessage {
    text: String,
    style: Style,
}
impl From<&RepoMessage> for Span<'_> {
    fn from(value: &RepoMessage) -> Self {
        Self::default()
            .content(value.text.clone())
            .style(value.style)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum RepoState {
    Identified {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
    },
    Configured {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
        branches: RepoBranches,
        log: Log,
    },
    Ready {
        repo_alias: RepoAlias,
        message: RepoMessage,
        alert: Option<String>,
        branches: RepoBranches,
        view_state: ViewState,
        main: Commit,
        next: Commit,
        dev: Commit,
        log: Log,
    },
}
impl RepoState {
    pub fn repo_alias(&self) -> &RepoAlias {
        match self {
            RepoState::Identified { repo_alias, .. }
            | RepoState::Configured { repo_alias, .. }
            | RepoState::Ready { repo_alias, .. } => repo_alias,
        }
    }

    #[tracing::instrument]
    pub fn update_branches(&mut self, branches: RepoBranches) {
        match self {
            Self::Configured {
                branches: state_branches,
                ..
            }
            | Self::Ready {
                branches: state_branches,
                ..
            } => {
                *state_branches = branches;
            }

            Self::Identified { .. } => (),
        }
    }

    #[tracing::instrument]
    pub fn update_log(&mut self, log: Log) {
        match self {
            Self::Configured { log: state_log, .. } | Self::Ready { log: state_log, .. } => {
                *state_log = log;
            }

            Self::Identified { .. } => {
                info!("git graph log ignored by ui");
            }
        }
    }

    #[tracing::instrument]
    pub fn update_message(&mut self, msg: impl Into<String> + std::fmt::Debug, colour: Color) {
        match self {
            Self::Identified { message, .. }
            | Self::Configured { message, .. }
            | Self::Ready { message, .. } => {
                info!(?msg, "updating ui");
                *message = RepoMessage::builder()
                    .text(msg.into())
                    .style(Style::default().fg(colour))
                    .build();
            }
        }
    }

    #[tracing::instrument]
    pub fn clear_alert(&mut self) {
        match self {
            Self::Identified { alert, .. }
            | Self::Configured { alert, .. }
            | Self::Ready { alert, .. } => {
                *alert = None;
            }
        }
    }

    #[tracing::instrument]
    pub fn alert(&mut self, msg: impl Into<String> + std::fmt::Debug) {
        let msg: String = msg.into();
        tracing::info!(%msg, "new tui alert");
        self.update_message("ALERT", Color::Red);
        match self {
            Self::Identified { alert, .. }
            | Self::Configured { alert, .. }
            | Self::Ready { alert, .. } => *alert = Some(msg),
        }
    }

    pub fn ready(self, main: Commit, next: Commit, dev: Commit) -> Self {
        match self {
            Self::Identified {
                repo_alias,
                message,
                alert,
            } => Self::Identified {
                repo_alias,
                message,
                alert,
            },
            Self::Configured {
                repo_alias,
                message,
                alert,
                branches,
                log,
            } => Self::Ready {
                repo_alias,
                message,
                alert,
                branches,
                view_state: ViewState::Expanded,
                main,
                next,
                dev,
                log,
            },
            Self::Ready {
                repo_alias,
                message,
                alert,
                branches,
                view_state,
                log,
                .. // drop existing main, next and dev to use parameters
            } => Self::Ready {
                repo_alias,
                message,
                alert,
                branches,
                view_state,
                main,
                next,
                dev,
                log,
            },
        }
    }
}

impl StatefulWidget for &State {
    type State = ScrollViewState;
    fn render(self, area: Rect, buf: &mut Buffer, state: &mut Self::State)
    where
        Self: Sized,
    {
        let block = Block::bordered()
            .title_top(
                Line::from(
                    if self.filter.is_empty() && self.key_focus == KeyFocus::None {
                        s!("")
                    } else {
                        format!(" Filter: {} ", self.filter)
                    },
                )
                .alignment(Alignment::Left)
                .style(if self.key_focus == KeyFocus::Filter {
                    Color::Red
                } else {
                    Color::Green
                }),
            )
            .title_top(
                Line::from(format!(" Git-Next v{} ", clap::crate_version!()).bold())
                    .alignment(Alignment::Center),
            )
            .title_bottom(
                Line::from(if self.key_focus == KeyFocus::Filter {
                    s!(" [esc] clear [tab/enter] finish ")
                } else {
                    s!("")
                })
                .alignment(Alignment::Left),
            )
            .title_bottom(
                Line::from(vec![
                    " [q]uit [/] filter ".into(),
                    self.beating_heart().into(),
                    " ".into(),
                ])
                .alignment(Alignment::Center),
            )
            .title_bottom(Line::from(format!(" {} ", time())).alignment(Alignment::Right))
            .border_set(border::THICK);
        let interior = block.inner(area);
        block.render(area, buf);
        match &self.mode {
            ServerState::Initial { tick } => Paragraph::new(format!("Loading...{tick}"))
                .centered()
                .render(interior, buf),
            ServerState::Configured { forges } => {
                ConfiguredAppWidget {
                    forges,
                    filter: &self.filter,
                }
                .render(interior, buf, state);
            }
        }
    }
}
